#include "Map2Rect.h"
#include "../Core_Mesh/Iterators.h"
#include "../CG_SPL/cg_spl.h"
#include <iostream>
#include <hash_map>
#include <fstream>

using namespace XMeshLib;

void Map2Rect::ComputeCornerAngle()
{// every face has three corners, compute and store their angles on corresponding halfedges
	for(MeshHalfedgeIterator heiter(cmesh); !heiter.end(); ++heiter )
	{
		Halfedge * he = * heiter;
		Point &p0 = he->target()->point(); 
		Point &p1 = he->he_next()->target()->point();
		Point &p2 = he->source()->point(); 
		double a = (p1 - p0).norm();
		double b = (p0 - p2).norm();
		double c = (p2 - p1).norm();
		double costheta = (a*a+b*b-c*c)/(2*a*b);
		double theta = acos(costheta);
		assert(theta >= 0 && theta <= 3.1415927);
		HEAngles[he] = theta;
	}
	HEAngles.rehash(HEAngles.size());
}

void Map2Rect::SetHarmonicWeights()
{
	ComputeCornerAngle();
	for(MeshEdgeIterator eiter(cmesh); !eiter.end(); ++eiter)
	{		
		Edge * e = *eiter;
		if (e->boundary())
		{// no need to consider these edges
			continue;
		}
		else
		{
			Halfedge * he1 = e->he(0)->he_next();
			Halfedge * he2 = e->he(1)->he_next();
			double weight = ( 1 /  tan(HEAngles[he1]) + 1 / tan(HEAngles[he2]) ) / 2; 
			EWeights[e] = weight;
		}
	}
	EWeights.rehash(EWeights.size());
	for (MeshVertexIterator vit(cmesh); !vit.end();++vit)
	{
		Vertex * v= *vit;
		if (v->boundary())
			continue;
		double wsum=0;
		for (VertexEdgeIterator veit(v); !veit.end();++veit)
		{
			Edge * e = *veit;
			wsum += EWeights[e];
		}
		WeightSums[v]=wsum;
	}
	WeightSums.rehash(WeightSums.size());
}

void Map2Rect::Parameterize(double u0, double v0, double u1, double v1, double u2, double v2, double u3, double v3, 
							const char outfile[], Loop & bLoop, double loopLength)			
{
	std::cout << "(1) Setting Corner UV" << std::endl;
	_1_SetCornerUV(u0, v0, u1, v1, u2, v2, u3, v3);
	std::cout << "(2) Computing Weights" << std::endl;
	_2_PrepareWeights();
	std::cout << "(3) Computing Boundary UV" << std::endl;	
	_3_ParameterizeBoundaryUV(bLoop, loopLength);
	std::cout << "(4) Computing Interior UV" << std::endl;
	_4_SolveInteriorUV();
	std::cout << "(5) Writing the Result" << std::endl;	
	_5_WriteMesh(outfile);
}

void Map2Rect::_1_SetCornerUV(double u0, double v0, double u1, double v1, double u2, double v2, double u3, double v3)
{
	texU[cornerV[0]]=u0;
	texU[cornerV[1]]=u1;
	texU[cornerV[2]]=u2;
	texU[cornerV[3]]=u3;
	texV[cornerV[0]]=v0;
	texV[cornerV[1]]=v1;
	texV[cornerV[2]]=v2;
	texV[cornerV[3]]=v3;
}

void Map2Rect::_2_PrepareWeights()
{	
	SetHarmonicWeights();
}


void Map2Rect::_3_ParameterizeBoundaryUV(Loop & bLoop, double loopLength)
{
	int bcount=4;
	//(2) Set Other Boundary UV
	int cornerInd=0;

	double accumLen=0;
	double aQuadLen = 0.25 * loopLength;
	for (int i=0;i<bLoop.lenth();++i)
	{//(2.1)get boundaries
		Vertex * srcV = cornerV[cornerInd]; 
		int cornerIndp1=(cornerInd+1)%4;
		Vertex * trgV = cornerV[cornerIndp1];
		Halfedge * he = bLoop.he(i);
		double clen = (he->target()->point()-he->source()->point()).norm();
		accumLen += clen;
		if (he->vertex()==cornerV[cornerIndp1])
		{
			cornerInd++;
			continue;
		}
		double alpha = (accumLen - (0.25*cornerInd*loopLength)) / aQuadLen;
		assert(alpha>=0 && alpha<=1);
		texU[he->vertex()] = texU[cornerV[cornerInd]] * (1-alpha) + texU[cornerV[cornerIndp1]] * alpha;
		texV[he->vertex()] = texV[cornerV[cornerInd]] * (1-alpha) + texV[cornerV[cornerIndp1]] * alpha;
		bcount++;
	}
	assert(bcount==bLoop.helist.size());
}

void Map2Rect::_4_SolveInteriorUV()
{		
	stdext::hash_map<int, int> mapVID2MInd;
	stdext::hash_map<int, int> mapMInd2VID;

	int inVSize=0;
	int coefSize=0;
	MeshVertexIterator viter(cmesh);
	for (; !viter.end(); ++viter)
	{
		Vertex * v=*viter;
		if (!v->boundary())
		{
			mapVID2MInd[v->id()]=inVSize;
			mapMInd2VID[inVSize]=v->id();
			++inVSize;
			++coefSize;
			texU[v]=0.5;
			texV[v]=0.5;
			for (VertexVertexIterator vviter(v); !vviter.end(); ++vviter)
			{
				Vertex * vv=*vviter;
				if (!vv->boundary())
					++coefSize;
			}
		}
	}

	mapVID2MInd.rehash(mapVID2MInd.size());
	mapMInd2VID.rehash(mapMInd2VID.size());

	//Solve M*X_u=vectA; M*X_v=vectB
	VECTOR_double x_u = VECTOR_double(inVSize, 0.0);
	VECTOR_double x_v = VECTOR_double(inVSize, 0.0);
	VECTOR_double vectA = VECTOR_double(inVSize, 0.0);
	VECTOR_double vectB = VECTOR_double(inVSize, 0.0);

	double *val= new double[coefSize];
	int *colind = new int[coefSize];
	int *rowind = new int[coefSize];
	
	int coefInd=0;
	int diagInd=0;
	for (viter.reset(); !viter.end(); ++viter)
	{
		Vertex * v = *viter;
		if (v->boundary())
			continue;
		int v1ind = mapVID2MInd[v->id()];

		rowind[coefInd]=v1ind;
		colind[coefInd]=v1ind;
		val[coefInd]=WeightSums[v];
		++coefInd;
		++diagInd;

		double rValueA, rValueB;
		rValueA=rValueB=0;
		for (VertexVertexIterator vviter(v); !vviter.end(); ++vviter)
		{
			Vertex * vv = *vviter;
			Edge * e = cmesh->vertexEdge(v,vv);
			double cweight = EWeights[e];
			
			if (vv->boundary())
			{
				rValueA += texU[vv] * cweight;
				rValueB += texV[vv] * cweight;
				continue;
			}
			int v2ind = mapVID2MInd[vv->id()];
			rowind[coefInd]=v1ind;
			colind[coefInd]=v2ind;
			val[coefInd]=-cweight;
			++coefInd;
		}
		vectA.p_[v1ind]=rValueA;
		vectB.p_[v1ind]=rValueB;
	}
	assert(coefInd==coefSize);
	assert(diagInd==WeightSums.size());

	Coord_Mat_double mat = Coord_Mat_double(inVSize, inVSize, coefSize, val, rowind, colind);
	double epsilon = 1e-13;
	int maxit = 10000;
	//solving linear systems mat * x_u = vectA, and mat * x_v = vectB;
	CG_SPL cg0 = CG_SPL(&mat, &vectA, &x_u, epsilon, maxit);
	CG_SPL cg1 = CG_SPL(&mat, &vectB, &x_v, epsilon, maxit);	
	cg0.solve_GMRES("x.info");
	cg1.solve_GMRES("y.info");
	for (int i=0;i<inVSize;++i)
	{
		Vertex * ver = cmesh->idVertex(mapMInd2VID[i]);
		assert(!ver->boundary());
		texU[ver] = (cg0.x)->p_[i]; 
		texV[ver] = (cg1.x)->p_[i];		
	}	
	delete [] val;
	delete [] colind;
	delete [] rowind;
}

void Map2Rect::_5_WriteMesh(const char file[])
{
	std::string keyuv = "uv";
	for (MeshVertexIterator viter(cmesh); !viter.end();++viter)
	{
		Vertex *ver = *viter;
		char keystr[100];
		sprintf(keystr, "uv=(%f %f)",texU[ver], texV[ver]);
		ver->PropertyStr()=keystr;
	}
	cmesh->write(file);
}
